#include "GrafoPonderado.h"

/* Inicializa estrutura bsica do grafo */
void criarGrafo(Grafo *G, int nVertices, int *erro) {
     
     if (nVertices > MAX_N_VERTICES) *erro = 1;

     else {
          
          *erro = 0;
          
          G->nVertices = nVertices;
          
          int i;          
          for (i = 0; i < G->nVertices; i++) {
                            
              G->LAdj[i].inicio = NULL;
              
          }
     }
}

/* Redefine o nmero de vrtices do grafo*/
void setNumVertices(Grafo* G, int nVertices) {

     G->nVertices = nVertices;
}

/* Retorna o nmero de vrtices do grafo*/
int getNumVertices(Grafo* G) {
 
    return G->nVertices;   
}


/* Insere uma aresta ponderada de v a u no final da lista de adjacncias de v, e vice-versa*/
void inserirAresta(Grafo *G, int v, int u, float peso, int *erro) {
     
     if (v >= G->nVertices || u >= G->nVertices) *erro = 1;
     
     else {
          
         *erro = 0;
          
         NoLista *p;
               
         p=(NoLista*) malloc(sizeof(NoLista));
         p->id = u;
		 p->peso = peso;       
         p->prox = NULL;
         
         if (G->LAdj[v].inicio == NULL)
         
            G->LAdj[v].inicio = p;
         
         else {
              
             NoLista *aux = G->LAdj[v].inicio;
             
             while (aux->prox != NULL)
                   aux = aux->prox;
                   
             aux->prox = p;
         }
         
         p=(NoLista*) malloc(sizeof(NoLista));
         p->id = v; 
		 p->peso = peso;              
         p->prox = NULL;
         
         if (G->LAdj[u].inicio == NULL)
         
            G->LAdj[u].inicio = p;
         
         else {
              
             NoLista *aux = G->LAdj[u].inicio;
             
             while (aux->prox != NULL)
                   aux = aux->prox;
                   
             aux->prox = p;
         }
         
           
     }
}

/* Verifica se a aresta de v a u est definida no grafo */
int existeAresta(Grafo *G, int v, int u, int *erro) {
    
    if (v >= G->nVertices || u >= G->nVertices) {
       
       *erro = 1;
       return 0; 
       
    } else {
         
         *erro = 0;
         
         int encontrou=0;
         NoLista *aux = G->LAdj[v].inicio;
         
         while (aux != NULL && !encontrou)
               
             if (aux->id == u) encontrou=1;
             else aux = aux->prox;
             
         return encontrou;
    }
}

/* Calcula o grau de um vrtice */
int grauVertice(Grafo *G, int v, int *erro) {
    
    int grau;
    
    if (v >= G->nVertices) *erro = 1;
    
    else {
          
          *erro = 0;
          
          grau = 0;
          NoLista *aux = G->LAdj[v].inicio;
          
          while (aux != NULL) {
                
                grau++;
                aux = aux->prox;
                
          }
    }
        
    return grau;
}

/* Imprime a lista de adjacncias do grafo*/
void imprimirGrafo(Grafo *G){
     
     int v;
     
     for (v = 0; v < G->nVertices; v++) {
     
          printf("%d:", v);
         
          NoLista *aux = G->LAdj[v].inicio;
         
          while (aux != NULL) {
                
              printf(" %d(%.2f)", aux->id, aux->peso);
              aux = aux->prox;
                
          }
          
          printf("\n");
     }
}

/* BFS a partir dd vrtice s */
void bfs(Grafo *G, int s, int dista[]) {

    int cor[G->nVertices];
    int inicioProc[G->nVertices];
    int fimProc[G->nVertices];
    int pai[G->nVertices];
//    int dista[G->nVertices];
    int timestamp = 0;
    
    int v, u;
    
    for (v = 0; v < G->nVertices; v++) {
        cor[v] = BRANCO;
        inicioProc[v] = NULO;
        fimProc[v] = NULO;
        pai[v] = NULO;
    }
    
    Fila F; 
    criarFila(&F);
    
    inserirFila(&F, s);
    
    dista[s] = 0;
    cor[s] = CINZA;
    inicioProc[s] = timestamp++;
//    printf("CINZA: %d %d\n", s, timestamp);        
    
    NoLista *viz;
     
    while (!filaVazia(&F)) {

//        imprimirFila(&F);
        v = removerFila(&F);
        
        viz = G->LAdj[v].inicio;
     
        while (viz != NULL) {
              
            u = viz->id;
     
            if (cor[u] == BRANCO) {
                
                cor[u] = CINZA;
                inicioProc[u] = timestamp++;
//                printf("CINZA: %d %d\n", u, timestamp); 
                
                pai[u] = v;
                dista[u] = dista[v] + 1;
                
                inserirFila(&F, u);
//                imprimirFila(&F);
                
            }

            viz = viz->prox;
            
        }
        
        cor[v] = PRETO;
        fimProc[v] = timestamp++;
//        printf("PRETO: %sd %d\n", v, timestamp);        
                     
    }
    
}


void dfs_(Grafo *G) {

    int pai[G->nVertices];

    int cor[G->nVertices];
    int inicioProc[G->nVertices];
    int fimProc[G->nVertices];
    int timestamp = 0;
    
    int v;
    
    for (v = 0; v < G->nVertices; v++) {

        pai[v] = NULO;		
        cor[v] = BRANCO;
        inicioProc[v] = NULO;
        fimProc[v] = NULO;

    }
    
    Pilha P; 
    criarPilha(&P);
    
    NoLista *viz;
    
	for (v = 0; v < G->nVertices; v++) {
    
    	/* Inicia/continua a travessia no primeiro vrtice ainda no processado */
		if (cor[v] == BRANCO) {
        
           	empilhar(&P, v);
    
           	while (!pilhaVazia(&P)) {
                 
//                imprimirPilha(&P); 
            	v = getTopo(&P);
            	
				
				if (cor[v] == BRANCO) {
				/* Se v est no topo da lista e branco, seu processamento deve ser iniciado */	
            	
            		cor[v] = CINZA;
            		inicioProc[v] = timestamp++;
					printf("CINZA: %d %d\n", v, timestamp);

				
					/* Percorre todos vrtices adjacentes a v */
        			viz = G->LAdj[v].inicio; 

            		while (viz != NULL) {
              			
              			/* Se ainda no iniciou o processamento do vertice adjacente, empilha */	
						if (cor[viz->id] == BRANCO) {
									          					
							empilhar(&P, viz->id);
							
							/* E adiciona-o  rvore de travessia */
							pai[viz->id] = v;
						
						} 

						viz = viz->prox;

					}						
				} else {
				
					if (cor[v] == CINZA) {
					/* Se v est no topo da lista e cinza, todas suas adjacncias j foram processadas e seu processamento pode ser encerrado */
					                    	
						cor[v] = PRETO;
						fimProc[v] = timestamp++;
						printf("PRETO: %d %d\n", v, timestamp);
					
					}
					
					/* Cinza ou Preto, o vrtice precisa ser desempilhado */
					desempilhar(&P);
					
				}
			}
		}    
	}  
}

void dfs(Grafo *G, int v, int pai[], int cor[], int inicioProc[], int fimProc[], int *timestamp) {
     
     cor[v] = CINZA;
     inicioProc[v] = (*timestamp)++;
     printf("CINZA: %d %d\n", v, *timestamp);
          
     NoLista *viz;
     viz = G->LAdj[v].inicio;
     
     while (viz != NULL) {
     
         if (cor[viz->id] == BRANCO) {
          
             pai[viz->id] = v;
             dfs(G, viz->id, pai, cor, inicioProc, fimProc, timestamp);
         }       
                
         viz = viz->prox;       
     }
     
     cor[v] = PRETO;
     fimProc[v] = (*timestamp)++;
     printf("PRETO: %d %d\n", v, *timestamp);
     
}


void iniciarDfs(Grafo *G) {

    int pai[G->nVertices];
    int cor[G->nVertices];
    int inicioProc[G->nVertices];
    int fimProc[G->nVertices];
    int timestamp = 0;
    
    int v;
    
    for (v = 0; v < G->nVertices; v++) {
		
        pai[v] = NULO;		
        cor[v] = BRANCO;
        inicioProc[v] = NULO;
        fimProc[v] = NULO;
        
    }
    
    for (v = 0; v < G->nVertices; v++)
        
        if (cor[v] == BRANCO)
        
           dfs(G, v, pai, cor, inicioProc, fimProc, &timestamp);
    
}


void classificarArestas_(Grafo *G, char matAdj[][MAX_N_VERTICES]) {

    int pai[G->nVertices];

    int cor[G->nVertices];
    int inicioProc[G->nVertices];
    int fimProc[G->nVertices];
    int timestamp = 0;
    
    int v;
    
    for (v = 0; v < G->nVertices; v++) {

        pai[v] = NULO;		
        cor[v] = BRANCO;
        inicioProc[v] = NULO;
        fimProc[v] = NULO;

    }
    
    Pilha P; 
    criarPilha(&P);
    
    NoLista *viz;
    
	for (v = 0; v < G->nVertices; v++) {
    
    	/* Inicia/continua a travessia no primeiro vrtice ainda no processado */
		if (cor[v] == BRANCO) {
        
           	empilhar(&P, v);
    
           	while (!pilhaVazia(&P)) {
                 
                imprimirPilha(&P); 
            	v = getTopo(&P);
//				v = desempilhar(&P);
            	
            	
				
				if (cor[v] == BRANCO) {
				/* Se v est no topo da lista e branco, seu processamento deve ser iniciado */	
            	
            		cor[v] = CINZA;
            		inicioProc[v] = timestamp++;
					printf("CINZA: %d %d\n", v, timestamp);

				
					/* Percorre todos vrtices adjacentes a v */
        			viz = G->LAdj[v].inicio; 

            		while (viz != NULL) {
              			
              			/* Se ainda no iniciou o processamento do vertice adjacente, empilha */	
						if (cor[viz->id] == BRANCO) {
									          					
							empilhar(&P, viz->id);
							
							/* E adiciona-o  rvore de travessia */
							pai[viz->id] = v;
						
							/* E classifica a aresta na classe T */						
							matAdj[v][viz->id] = 'T';
						
						} else if (cor[viz->id] == CINZA)
                      	/* Se j iniciou o processamento, classifica a aresta nas classes B, F ou C */	
						
							/* Retorno */
							matAdj[v][viz->id] = 'B';
					
						else if (cor[viz->id] == PRETO)
              
              				/* Avano */
							if (inicioProc[viz->id] > inicioProc[v])  matAdj[v][viz->id] = 'F';
              
              				/* Cruzamento */
			  				else matAdj[v][viz->id] = 'C';


						viz = viz->prox;

					}						
				} else {
				
					if (cor[v] == CINZA) {
					/* Se v est no topo da lista e cinza, todas suas adjacncias j foram processadas e seu processamento pode ser encerrado */
					                    	
						cor[v] = PRETO;
						fimProc[v] = timestamp++;
						printf("PRETO: %d %d\n", v, timestamp);
					
					}
					
					/* Cinza ou Preto, o vrtice precisa ser desempilhado */
					desempilhar(&P);
					
				}
			}
		}    
	}   					

}


void classificarArestas(Grafo *G, int v, int cor[], int inicioProc[], int *timestamp, char matAdj[][MAX_N_VERTICES]) {
     
    cor[v] = CINZA;
    inicioProc[v] = (*timestamp)++;
    printf("CINZA: %d %d\n", v, *timestamp);
          
    NoLista *viz;
	viz = G->LAdj[v].inicio;
     
    int u;
     
	while (viz != NULL) {
     
    	if (cor[viz->id] == BRANCO) {
          
        	matAdj[v][viz->id] = 'T';
            classificarArestas(G, viz->id, cor, inicioProc, timestamp, matAdj);
             
		} else if (cor[viz->id] == CINZA)
                
			matAdj[v][viz->id] = 'B';
	         		
		else if (cor[viz->id] == PRETO)
              
			if (inicioProc[viz->id] > inicioProc[v])  matAdj[v][viz->id] = 'F';
              
			else matAdj[v][viz->id] = 'C';
         
         viz = viz->prox;       
	}
     
	cor[v] = PRETO;
//     printf("PRETO: %d %d\n", v, *timestamp);
     
}


void iniciarClassificarArestas(Grafo *G, char matAdj[][MAX_N_VERTICES]) {
      
    int cor[G->nVertices];
    int inicioProc[G->nVertices];
    int timestamp = 0;
    
    int v;
    
    for (v = 0; v < G->nVertices; v++) {
        cor[v] = BRANCO;
        inicioProc[v] = NULO;        
    }
    
    for (v = 0; v < G->nVertices; v++)
    
        if (cor[v] == BRANCO)
        
           classificarArestas(G, v, cor, inicioProc, &timestamp, matAdj);
}

/* AGM obtida pelo Prim*/
void prim(Grafo *G, int pivo, int pai[]) {
	
	int usado[G->nVertices];
	
	NoHeap heap[MAX_N_ARESTAS];
	int tamHeap = 0;
    
    int u, v;
    int achou;
    
    for (v = 0; v < G->nVertices; v++) {

        pai[v] = NULO;		
        usado[v] = 0;

    }
    
	NoLista *viz;
	
	do {
    
    	/* Dado um vrtice piv, especula todas as arestas a vrtices adjacentes */
        viz = G->LAdj[pivo].inicio;
    
        while (viz != NULL) {
    
            if (!usado[viz->id]) {
                       
                heap[tamHeap].pivo = pivo;
                heap[tamHeap].vAdj = viz->id;
                heap[tamHeap].val = viz->peso;
        
                tamHeap ++;
            }   
        
            viz = viz->prox;
       	}
    
       	/* Ordena as arestas de acordo com o peso */    
       	minHeapSort(tamHeap, heap);

       	/* Seleciona aresta com menor peso cujo vertice adjacente ainda no esta na AGM */
       	achou = 0;
    	while (tamHeap > 0 && !achou) {
    
    		u = heap[tamHeap-1].pivo;
			v = heap[tamHeap-1].vAdj;
    
           	if (!usado[v]) {
				            
               /* pe na AGM */
				pai[v] = u;              
               	pivo = v;
               	usado[pivo] = 1;
               	
            	achou = 1;
           	}
        
           	tamHeap --;      
       	}
    
    } while (tamHeap > 0);
    	
}
